import math
from collections import deque

import numpy as np

from marlenv.core.grid_util import rgb_from_grid
from marlenv.core.snake import Cell, CellColors
from .snake_env import SnakeEnv

"""
    UP = (-1, 0)
    RIGHT = (0, 1)
    DOWN = (1, 0)
    LEFT = (0, -1)
"""


class GraphSnakeEnv(SnakeEnv):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _init_obs(self):
        if self.image_obs:
            self.obs = deque(maxlen=self.frame_stack)
            for _ in range(self.frame_stack):
                self.obs.append(rgb_from_grid(self.grid, Cell, CellColors))
            obs = [np.concatenate(list(self.obs), axis=-1)
                   for _ in range(self.num_snakes)]
        else:
            self.obs = deque(maxlen=self.frame_stack)
            _obs = self._encode(self.grid, vision_range=self.vision_range)
            for _ in range(self.frame_stack):
                self.obs.append(_obs)
            obs = list(zip(*list(self.obs)))
            obs = [np.concatenate(o, axis=-1) for o in obs]

        proc_obs = self._process_obs(obs)

        return proc_obs

    def _get_obs(self):
        obs = super()._get_obs()
        obs = self._process_obs(obs)

        return obs

    def _process_obs(self, obs):

        if not self.observer == 'snake':
            raise ValueError(
                "This is not yet implemented for 'human' observers.")
        if self.image_obs:
            raise ValueError(
                "This is not yet implemented for 'image' observation.")

        vision_range = 5  # range of vision in default five
        if self.vision_range:
            vision_range = self.vision_range
        proc_obs = []
        sqrt2 = math.sqrt(2)
        snake_idx = 0
        for snake in self.snakes:  # for each snake
            if not snake.alive:
                continue
            proc_ob = []
            angle = math.atan2(snake.direction.value[1],
                               snake.direction.value[0])
            head = snake.head_coord
            if self.vision_range:  # if so, the head is at the center
                head = (self.vision_range, self.vision_range)
            for l in range(3):  # for each of three directions except backward
                dx = (int(math.cos(angle + self.action_dict[l])),
                      int(math.sin(angle + self.action_dict[l])))
                proc_ob.append(np.zeros((self.obs_ch,)))
                for i in range(vision_range):
                    temp_ob = obs[snake_idx][head[0] + dx[0] * (i + 1)][
                        head[1] + dx[1] * (i + 1)]
                    proc_ob[-1] += temp_ob / (i + 1)
                    if temp_ob[0] == 1:  # up to the wall
                        break
            for l in [(0, 1), (0, 2)]:  # each of two diagonal directions
                dx = [(int(math.cos(angle + self.action_dict[l[q]])),
                       int(math.sin(angle + self.action_dict[l[q]]))) for q in
                      range(2)]
                proc_ob.append(np.zeros((self.obs_ch,)))
                for i in range(vision_range):
                    temp_ob = \
                    obs[snake_idx][head[0] + (dx[0][0] + dx[1][0]) * (i + 1)][
                        head[1] + (dx[0][1] + dx[1][1]) * (i + 1)]
                    if temp_ob[0] == 1:
                        proc_ob[-1] += temp_ob / ((i + 1) * sqrt2)
                        break
                    proc_ob[-1] += temp_ob / ((i + 1) * sqrt2)
            proc_obs.append(np.array(proc_ob))
            snake_idx += 1

        return np.array(proc_obs)



from typing import List

import numpy as np
import copy
from PIL import Image

DOWN = (0, 1)
RIGHT = (1, 0,)
UP = (0, -1)
LEFT = (-1, 0)
SHIFTS = [DOWN, RIGHT, UP, LEFT]


def make_grid(height, width, empty_value=0, wall_value=1) -> np.ndarray:
    grid = np.full((height, width), fill_value=empty_value)
    # construct wall
    grid[[0, -1]] = wall_value
    grid[:, [0, -1]] = wall_value

    return grid


def make_grid_from_txt(map_path: str, mapper: dict) -> np.ndarray:
    with open(map_path, 'r') as fp:
        lines = fp.read().split('\n')

    ls = []
    for line in lines:
        # split string into list of char
        ls.append([mapper[c] for c in list(line)])
    grid = np.asarray(ls)

    return grid


def find_k_consec(grid: np.ndarray, k: int):
    # Returns list of possible k consecutive cells from given grid
    empty_mask = (grid == 0)
    answers = []
    for r in range(grid.shape[0]):
        start = None
        end = None
        for rc in range(grid.shape[1]):
            if empty_mask[r, rc]:
                if start is None:
                    start = rc
                else:
                    end = rc
                if end - start == k + 1:
                    answers.append([(r, start), (r, end)])
                    start += 1
            else:
                start = None
                end = None
    for c in range(grid.shape[1]):
        start = None
        end = None
        for cr in range(grid.shape[0]):
            if empty_mask[r, rc]:
                if start is None:
                    start = cr
                else:
                    end = cr
                if end - start == k + 1:
                    answers.append([(start, c), (end, c)])
                    start += 1
            else:
                start = None
                end = None
    return answers


def dfs_sweep_empty(grid: np.ndarray, k: int):
    answers = []
    empty_mask = grid == 0
    for r in range(grid.shape[0]):
        for c in range(grid.shape[1]):
            if empty_mask[(r, c)]:
                answers.extend(_dfs_helper(empty_mask, (r, c), [], k))
    return answers


def _dfs_helper(grid: np.ndarray, node: tuple, history: List[tuple], k: int):
    history.append(node)
    answers = []
    if len(history) == k:
        answers.append(history)
    else:
        for shift in SHIFTS:
            # Get shifted rows and colums
            rs, cs = node[0] + shift[0], node[1] + shift[1]
            candidate = (rs, cs)
            if (_inbound(candidate, grid)
               and candidate not in history and grid[candidate]):
                if not _head_blocked(grid, history, candidate):
                    answers.extend(
                        _dfs_helper(grid, candidate, copy.deepcopy(history), k)
                    )
    return answers


def _head_blocked(mask: np.ndarray, history: List[tuple], extra_node: tuple):
    check_status = 0
    first_node = history[0]
    for shift in SHIFTS:
        node = first_node[0] + shift[0], first_node[1] + shift[1]
        if (mask[node] == 0 or node in history or node == extra_node
           or not _inbound(node, mask)):
            check_status += 1
    return check_status == len(SHIFTS)


def _inbound(node, grid):
    r, c = node
    return r >= 0 and c >= 0 and r < grid.shape[0] and c < grid.shape[1]


def random_empty_coord(grid: np.ndarray):
    # has to sweep entire grid O(H*W)
    xs, ys = np.where(grid == 0)
    idx = np.random.randint(0, len(xs) - 1)

    return xs[idx], ys[idx]


def random_empty_coords(grid, num_coords: int):
    xs, ys = np.where(grid == 0)
    if len(xs) == 0:
        return None, None
    idxes = np.random.randint(0, len(xs), size=num_coords)
    # coords = np.stack(xs[idxes], ys[idxes]).T

    return xs[idxes], ys[idxes]


def poll_empty_coord(grid):
    h, w = grid.shape
    # max index - wall * 2
    x, y = 0, 0
    while grid[x, y] != 0:
        x = np.random.randint(0, h - 1)
        y = np.random.randint(0, w - 1)

    # +1 to omit wall
    return x, y


def draw(grid, coords: List[tuple], value: int):
    h, w = grid.shape
    xs, ys = [], []
    for p in coords:
        xs.append(p[0])
        ys.append(p[1])

    # wall
    if (0 in xs) or ((h - 1) in xs) or (0 in ys) or ((w - 1) in ys):
        return False

    grid[xs, ys] = value

    return True


def rgb_from_grid(grid, enum, color_dict):
    rgb_array = np.zeros((*grid.shape, 3), dtype=np.uint8)
    for r in range(grid.shape[0]):
        for c in range(grid.shape[1]):
            cell_value = grid[r, c] % 10
            cell_id = grid[r, c] // 10
            color_list = color_dict[enum(cell_value).value]
            cell_color = np.array(color_list[cell_id % len(color_list)])
            cycle = cell_id // len(color_list)
            rgb_array[r, c] = (cell_color * 0.7**cycle).astype(np.uint8)

    return rgb_array


def image_from_grid(grid, enum, color_dict, max_size=300):
    bigger = max(list(grid.shape))
    scale = max(max_size // bigger, 1)
    rgb_array = rgb_from_grid(grid, enum, color_dict)
    rgb_array = np.repeat(np.repeat(rgb_array, scale, axis=0), scale, axis=1)
    image = Image.fromarray(rgb_array, 'RGB')

    return image



from .snake_env import SnakeEnv


class CoopSnakeEnv(SnakeEnv):
    """
        Identical to SnakeEnv's step function except for the terminating
        condition. if all(dones) -> any(dones)
        This is for finishing an episode if there's at least one dead snake.
        step() returns done = [True] * num_snakes if there's a dead snake.
        """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def step(self, actions):
        obs, rews, dones, info = super().step(actions)
        if self._done_fn(dones):
            dones = [True] * self.num_snakes

        return obs, rews, dones, info

    def _done_fn(self, dones):
        return any(dones)


from collections import deque
from enum import Enum


class Cell(Enum):
    EMPTY = 0
    WALL = 1
    FRUIT = 2
    HEAD = 3
    BODY = 4
    TAIL = 5


_color_wheel = [(104, 255, 0), (255, 191, 0), (255, 0, 92), (0, 111, 255)]


CellColors = {
    Cell.EMPTY.value: [(0, 0, 0)],
    Cell.WALL.value: [(32, 32, 32)],
    Cell.FRUIT.value: [(223, 7, 22)],
    Cell.HEAD.value: _color_wheel,
    Cell.BODY.value: _color_wheel,
    Cell.TAIL.value: _color_wheel
}


class Direction(Enum):
    UP = (-1, 0)
    RIGHT = (0, 1)
    DOWN = (1, 0)
    LEFT = (0, -1)

    """
    little magic for convenience, coord + direction
    """

    def __radd__(self, other):
        dr, dc = self.value
        return other[0] + dr, other[1] + dc

    def __rsub__(self, other):
        dr, dc = self.value
        return other[0] - dr, other[1] - dc


class Snake:
    def __init__(self, idx, coords):
        assert len(coords) > 1
        self.idx: int = idx
        self.head_coord: tuple = coords[0]
        self.tail_coord: tuple = coords[-1]
        self.direction: Direction = Direction(
            (coords[0][0] - coords[1][0],
            coords[0][1] - coords[1][1])
        )
        prev_coord = self.head_coord
        direction_list = []
        for next_coord in coords[1:]:
            direction = Direction(
                (prev_coord[0] - next_coord[0], prev_coord[1] - next_coord[1])
            )
            direction_list.append(direction)
            prev_coord = next_coord

        self.directions = deque(direction_list)

        self.alive = True
        self._reset_reward_state()

    def __len__(self):
        return len(self.directions + 1)

    def _reset_reward_state(self):
        self.fruit = False
        self.death = False
        self.kills = 0
        self.win = False
        self.reward = 0.

    @property
    def coords(self):
        coord = self.head_coord
        coords = [coord]
        for direction in self.directions:
            coord -= direction
            coords.append(coord)

        return coords

    def move(self):
        self.head_coord += self.direction
        self.directions.appendleft(self.direction)

        prev_tail_coord = None
        if not self.fruit:
            prev_tail_coord = self.tail_coord
            tail_direction = self.directions.pop()
            self.tail_coord += tail_direction
        self._reset_reward_state()

        return prev_tail_coord



import io
import os
import shutil
import sys

import gym
import marlenv
import pytest
from PIL import Image


@pytest.fixture
def snake_env():
    custom_rew = {
        'fruit': 1.0,
        'kill': 2.0,
        'lose': 3.0,
        'win': 4.0,
        'time': 0.1,
    }
    env = gym.make('Snake-v1', num_fruits=4, num_snakes=1,
                   reward_dict=custom_rew, disable_env_checker=True)

    return env


@pytest.fixture
def coop_snake_env():
    custom_rew = {
        'fruit': 1.0,
        'kill': 2.0,
        'lose': 3.0,
        'win': 4.0,
        'time': 0.1,
    }
    env = gym.make('SnakeCoop-v1', num_fruits=4, num_snakes=1,
                   reward_dict=custom_rew, disable_env_checker=True)

    return env


@pytest.fixture
def graph_snake_env():
    custom_rew = {
        'fruit': 1.0,
        'kill': 2.0,
        'lose': 3.0,
        'win': 4.0,
        'time': 0.1,
    }
    env = gym.make('SnakeGraph-v1', num_fruits=4, num_snakes=1,
                   reward_dict=custom_rew, disable_env_checker=True)

    return env


def rollout(env, n=100, render_mode=None):
    num_snakes = env.num_snakes
    obs = env.reset()
    dones = [False] * num_snakes

    for _ in range(n):
        if not all(dones):
            if render_mode:
                env.unwrapped.render(render_mode)
            ac = [env.action_space.sample() for _ in range(num_snakes)]
            obs, rews, dones, _ = env.step(ac)


def test_snake_env_rollout(snake_env):
    n_rollouts = 100
    rollout(snake_env, n=n_rollouts)


def test_coop_snake_env_rollout(coop_snake_env):
    n_rollouts = 100
    rollout(coop_snake_env, n=n_rollouts)


def test_graph_env_rollout(graph_snake_env):
    n_rollouts = 100
    rollout(graph_snake_env, n=n_rollouts)


@pytest.fixture
def processed_snake_env(snake_env):
    n_rollouts = 100
    rollout(snake_env, n=n_rollouts, render_mode='gif')

    return snake_env


def test_save_gif_default(processed_snake_env):
    env = processed_snake_env
    image_dir = env.save_gif()
    assert os.path.exists(image_dir)

    gif = Image.open(image_dir)
    gif.seek(1)

    d = os.path.dirname(image_dir)
    shutil.rmtree(d)


def test_save_gif_fileobj(processed_snake_env):
    env = processed_snake_env

    with io.BytesIO() as fileobj:
        output = env.save_gif(fileobj)

    assert sys.getsizeof(output) > 0



import sys
from enum import Enum
import multiprocessing as mp
import numpy as np

import gym
from gym.vector.async_vector_env import AsyncVectorEnv
from gym.error import AlreadyPendingCallError, NoAsyncCallError
from gym.vector.utils import write_to_shared_memory


class AsyncState(Enum):
    DEFAULT = 'default'
    WAITING_RESET = 'reset'
    WAITING_STEP = 'step'
    WAITING_RENDER = 'render'


class SingleAgent(gym.Wrapper):
    def __init__(self, env):
        super().__init__(env)
        assert env.num_snakes == 1, "Number of player must be one"
        self.action_space = gym.spaces.Discrete(len(self.env.action_dict))
        if self.vision_range:
            h = w = self.vision_range * 2 + 1
            self.observation_space = gym.spaces.Box(
                self.low, self.high,
                shape=(h, w, self.obs_ch), dtype=np.uint8)  # 8
        else:
            self.observation_space = gym.spaces.Box(
                self.low, self.high,
                shape=(*self.grid_shape, self.obs_ch), dtype=np.uint8)  # 8

    def reset(self, **kwargs):
        wrapped_obs = self.env.reset(**kwargs)
        return wrapped_obs[0]

    def step(self, action, **kwargs):
        obs, rews, dones, infos = self.env.step([action], **kwargs)
        return obs[0], rews[0], dones[0], {}


class SingleMultiAgent(gym.Wrapper):
    def __init__(self, env):
        super().__init__(env)
        self.action_space = gym.spaces.Discrete(len(self.env.action_dict))
        if self.vision_range:
            h = w = self.vision_range * 2 + 1
            self.observation_space = gym.spaces.Box(
                self.low, self.high,
                shape=(self.num_snakes, h, w, self.obs_ch), dtype=np.uint8)
        else:
            self.observation_space = gym.spaces.Box(
                self.low, self.high,
                shape=(self.num_snakes, *self.grid_shape, self.obs_ch),
                dtype=np.uint8)


class AsyncVectorMultiEnv(AsyncVectorEnv):
    def __init__(self, env_fns, **kwargs):
        super().__init__(env_fns, worker=_worker_shared_memory, **kwargs)
        self.default_state = None

    def render_async(self):
        self._assert_is_running()
        if self._state.value != AsyncState.DEFAULT.value:
            raise AlreadyPendingCallError(
                'Calling `render_async` while waiting '
                'for a pending call to `{0}` to complete.'.format(
                    self._state.value), self._state.value)
        else:
            self.default_state = self._state
        self.parent_pipes[0].send(('render', None))
        self._state = AsyncState.WAITING_RENDER

    def render_wait(self, timeout=None):
        self._assert_is_running()
        if self._state.value != AsyncState.WAITING_RENDER.value:
            raise NoAsyncCallError(
                'Calling `render_wait` without any prior '
                'call to `render_async`.', AsyncState.WAITING_RESET.value)

        if not self._poll(timeout):
            self._state = self.default_state
            raise mp.TimeoutError(
                'The call to `render_wait` has timed out after '
                '{0} second{1}.'.format(timeout, 's' if timeout > 1 else ''))

        result, success = self.parent_pipes[0].recv()
        self._raise_if_errors([success])
        self._state = self.default_state

        return result

    def render(self, *args):
        self.render_async()
        return self.render_wait()


def _worker(index, env_fn, pipe, parent_pipe, shared_memory, error_queue):
    assert shared_memory is None
    env = env_fn()
    parent_pipe.close()
    try:
        while True:
            command, data = pipe.recv()
            if command == 'reset':
                observation = env.reset()
                pipe.send((observation, True))
            elif command == 'step':
                observation, reward, done, info = env.step(data)
                if type(done) == bool:
                    cond = done
                else:
                    cond = all(done)
                if cond:
                    observation = env.reset()
                pipe.send(((observation, reward, done, info), True))
            elif command == 'seed':
                env.seed(data)
                pipe.send((None, True))
            elif command == 'close':
                pipe.send((None, True))
                break
            elif command == 'render':
                img = env.render('rgb_array')
                pipe.send((img, True))
            elif command == '_check_observation_space':
                pipe.send((data == env.observation_space, True))
            else:
                raise RuntimeError(
                    'Received unknown command `{0}`. Must '
                    'be one of {`reset`, `step`, `seed`, `close`, '
                    '`_check_observation_space`}.'.format(command))
    except (KeyboardInterrupt, Exception):
        error_queue.put((index,) + sys.exc_info()[:2])
        pipe.send((None, False))
    finally:
        env.close()


def _worker_shared_memory(index, env_fn, pipe, parent_pipe, shared_memory,
                          error_queue):
    assert shared_memory is not None
    env = env_fn()
    observation_space = env.observation_space
    parent_pipe.close()
    try:
        while True:
            command, data = pipe.recv()
            if command == 'reset':
                observation = env.reset()
                write_to_shared_memory(index, observation, shared_memory,
                                       observation_space)
                pipe.send((None, True))
            elif command == 'step':
                observation, reward, done, info = env.step(data)
                if type(done) == bool:
                    cond = done
                else:
                    cond = all(done)
                if cond:
                    observation = env.reset()
                write_to_shared_memory(index, observation, shared_memory,
                                       observation_space)
                pipe.send(((None, reward, done, info), True))
            elif command == 'seed':
                env.seed(data)
                pipe.send((None, True))
            elif command == 'close':
                pipe.send((None, True))
                break
            elif command == 'render':
                img = env.render('rgb_array')
                pipe.send((img, True))
            elif command == '_check_observation_space':
                pipe.send((data == observation_space, True))
            else:
                raise RuntimeError(
                    'Received unknown command `{0}`. Must '
                    'be one of {`reset`, `step`, `seed`, `close`, '
                    '`_check_observation_space`}.'.format(command))
    except (KeyboardInterrupt, Exception):
        error_queue.put((index,) + sys.exc_info()[:2])
        pipe.send((None, False))
    finally:
        env.close()


def make_snake(num_envs=1, num_snakes=4, env_id="Snake-v1", **kwargs):
    """A function just for me.

    :param my_arg: The first of my arguments.
    :param my_other_arg: The second of my arguments.

    :returns: A message (just for me, of course).
    """
    if num_snakes > 1:
        env_wrapper = SingleMultiAgent
    else:
        env_wrapper = SingleAgent
    if num_envs > 1:
        vec_wrapper = AsyncVectorMultiEnv

    def _make():
        env = gym.make(env_id, num_snakes=num_snakes, **kwargs)
        env = env_wrapper(env)
        return env

    dummyenv = _make()
    observation_shape = dummyenv.observation_space.shape
    if num_snakes > 1:
        observation_shape = observation_shape[1:]
    action_shape = (dummyenv.action_space.n,)
    high = dummyenv.observation_space.high
    low = dummyenv.observation_space.low

    if 'Discrete' in str(type(dummyenv.action_space)):
        action_info = {'action_n': dummyenv.action_space.n}
        discrete = True

    if 'Box' in str(type(dummyenv.action_space)):
        action_info = {
            'action_high': dummyenv.action_space.high,
            'action_low': dummyenv.action_space.low
        }
        discrete = False

    del dummyenv

    if num_envs > 1:
        env = vec_wrapper([_make for _ in range(num_envs)])
    else:
        env = _make()

    properties = {
        'high': high,
        'low': low,
        'num_envs': num_envs,
        'num_snakes': num_snakes,
        'discrete': discrete,
        'action_info': action_info,
    }

    return env, observation_shape, action_shape, properties



from gym.envs.registration import register

register(
    id='Snake-v1',
    entry_point='marlenv.envs.snake_env:SnakeEnv',
)

register(
    id='SnakeGraph-v1',
    entry_point='marlenv.envs.graph_snake_env:GraphSnakeEnv',
)

register(
    id='SnakeCoop-v1',
    entry_point='marlenv.envs.coop_snake_env:CoopSnakeEnv'
)


from setuptools import setup, find_packages
import io

with open("README.md") as f:
    with io.open("README.md", encoding="utf-8") as f:
        long_description = f.read()
        
setup(
    name='marlenv',
    version='1.0.1',
    url='https://github.com/kc-ml2/marlenv',
    author='Tae Min Ha, Daniel Nam, Won Seok Jung',
    author_email='contact@kc-ml2.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        'gym==0.24.1',
    ],
    python_requires='>=3.8',

)



from collections import defaultdict, deque
from typing import List, Tuple

import os
import math
import datetime
import warnings

import numpy as np
import gym
from gym.utils import seeding

from marlenv.core.grid_util import (
    random_empty_coords, draw, make_grid, dfs_sweep_empty,
    rgb_from_grid, image_from_grid)
from marlenv.core.snake import Direction, Snake, Cell, CellColors
from marlenv.envs.constants import FEATURE_CHANNEL, RGB_CHANNEL


class SnakeEnv(gym.Env):
    default_action_dict = {
        'noop': 0,
        'left': 1,
        'right': 2,
        'down': 3,
        'up': 4
    }

    action_angle_dict = {
        0: 0.0,
        1: math.pi / 2.0,
        2: -math.pi / 2.0
    }

    default_reward_dict = {
        'fruit': 1.0,
        'kill': 0.,
        'lose': 0.0,
        'win': 0.,
        'time': 0.,
    }
    reward_keys = default_reward_dict.keys()

    max_episode_steps = 1e4

    def __init__(
            self,
            height=20,
            width=20,
            num_snakes=4,
            snake_length=3,
            vision_range=None,
            frame_stack=1,
            observer='snake',
            # if 'snake' three actions, if 'human' five actions
            *args,
            **kwargs
    ):
        """
        kwargs
        'reward_dict', 'num_fruits'
        """

        reward_dict = kwargs.pop('reward_dict', SnakeEnv.default_reward_dict)
        if reward_dict.keys() != SnakeEnv.reward_keys:
            raise KeyError(
                f'reward dict keys must correspond to {SnakeEnv.reward_keys}'
            )
        else:
            self.reward_dict = reward_dict
        self.max_episode_steps = kwargs.pop('max_episode_steps',
                                            SnakeEnv.max_episode_steps)

        self.num_snakes = num_snakes
        self.num_fruits = kwargs.pop('num_fruits',
                                     int(round(num_snakes * 0.8)))
        self.grid_shape: Tuple = (height, width)
        self.grid: np.ndarray
        self.snakes: List[Snake]
        self.snake_length = snake_length
        self.vision_range = vision_range
        self.observer = observer

        self.low = 0
        self.image_obs = False
        if self.image_obs:
            self.high = 255
        else:
            self.high = 1
        if self.observer == 'human':
            self.action_dict = SnakeEnv.default_action_dict
        elif self.observer == 'snake':
            self.action_dict = SnakeEnv.action_angle_dict

        self.action_space = gym.spaces.Discrete(
            len(self.action_dict) * self.num_snakes
        )

        self.frame_stack = frame_stack
        default_ch = RGB_CHANNEL if self.image_obs else FEATURE_CHANNEL
        self.obs_ch = default_ch * self.frame_stack

        if self.vision_range:
            h = w = self.vision_range * 2 + 1
            self.observation_space = gym.spaces.Box(
                self.low,
                self.high,
                shape=[self.num_snakes, h, w, self.obs_ch],
                dtype=np.uint8
            )
        else:
            self.observation_space = gym.spaces.Box(
                self.low,
                self.high,
                shape=[self.num_snakes, *self.grid_shape, self.obs_ch],
                dtype=np.uint8
            )

    def reset(self):
        # Create the grid base
        self.grid = make_grid(*self.grid_shape,
                              empty_value=Cell.EMPTY.value,
                              wall_value=Cell.WALL.value)
        # Generate and add snake to the grid
        self.snakes = self._generate_snakes()
        for snake in self.snakes:
            coords = snake.coords
            snake_id = 10 * snake.idx
            for coord in coords:
                self.grid[coord] = Cell.BODY.value + snake_id
            self.grid[snake.head_coord] = Cell.HEAD.value + snake_id
            self.grid[snake.tail_coord] = Cell.TAIL.value + snake_id

        # Generate fruit and add to the grid
        xs, ys = self._generate_fruits(self.num_fruits)
        self.grid[xs, ys] = Cell.FRUIT.value

        self.alive_snakes = self.num_snakes
        self.frame_buffer = []

        obs = self._init_obs()

        # Episodic stats
        self._reset_epi_stats()
        self.episode_length = 0

        return np.array(obs, dtype=np.uint8)

    def seed(self, seed=42):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='ascii'):
        if mode == 'ascii':
            # Just for debugging
            SYM2CHR = {
                Cell.EMPTY.value: '.',
                Cell.WALL.value: '#',
                Cell.FRUIT.value: 'o',
                Cell.BODY.value: 'b',
                Cell.HEAD.value: 'H',
                Cell.TAIL.value: 't'
            }

            # CHR2SYM = {v: k for k, v in SYM2CHR.items()}

            data = np.vectorize(SYM2CHR.get)(self.grid % 10)
            data = [''.join(i) for i in data]
            data = '\n'.join(data)
            print(data)
        elif mode == 'gif':
            # Save game play as gif
            # Create a game scene from the grid
            # Append to buffer
            game_frame = image_from_grid(self.grid, Cell, CellColors)
            self.frame_buffer.append(game_frame)
        elif mode == 'rgb_array':
            rgb_array = rgb_from_grid(self.grid, Cell, CellColors)
            return rgb_array
        elif mode == 'human':
            # Run pygame
            pass

    def close(self):
        pass

    def step(self, actions):
        """
        Time Complexity:
            O(3*num_snakes)
            + (O(H*W) w/ fruits(maybe better approach, like random polling))

        updates snake status and grid
        all snake statuses are updated only in step method
        """

        if isinstance(actions, int):
            actions = [actions]
        assert len(actions) == self.num_snakes
        for i, ac in enumerate(actions):
            if isinstance(ac, np.ndarray):
                actions[i] = ac.item()
        # preprocess
        next_head_coords = defaultdict(list)
        alive_snakes = []
        for snake, action in zip(self.snakes, actions):
            if snake.alive:
                if self.observer == 'human':
                    snake.direction = self._next_direction_global(
                        snake.direction, action)
                elif self.observer == 'snake':
                    snake.direction = self._next_direction(snake.direction,
                                                           action)
                new_head_coord = snake.head_coord + snake.direction
                next_head_coords[new_head_coord].append(snake.idx)
                alive_snakes.append(snake.idx)
        dead_idxes, fruit_idxes, fruit_taken = self._check_collision(
            next_head_coords)

        self.alive_snakes -= len(dead_idxes)
        for idx in dead_idxes:
            self.snakes[idx].death = True
            self.snakes[idx].alive = False
        for idx in fruit_idxes:
            tail_coord = self.snakes[idx].tail_coord
            if tail_coord in next_head_coords.keys():
                for di in next_head_coords[tail_coord]:
                    self.snakes[di].death = True
                    self.snakes[di].alive = False
                    self.alive_snakes -= 1
                    self.snakes[idx].kills += 1
            self.snakes[idx].fruit = True
        if self.alive_snakes == 1 and self.num_snakes > 1:
            for snake in self.snakes:
                if snake.alive:
                    # print('player {} wins!'.format(snake.idx))
                    snake.win = True
                    break

        rews = []
        dones = []
        fruits, kills = [], []
        # Update snake rews, stats, etc., and update the grid accordingly
        for snake in self.snakes:
            if not snake.death and not snake.alive:
                snake.reward = 0.
                rews.append(snake.reward)
                fruits.append(0)
                kills.append(0)
            else:
                snake.reward = self.reward_dict['time'] * snake.alive
                snake.reward += self.reward_dict['fruit'] * snake.fruit
                snake.reward += self.reward_dict['lose'] * snake.death
                snake.reward += self.reward_dict['kill'] * snake.kills
                snake.reward += self.reward_dict['win'] * snake.win
                rews.append(snake.reward)
                fruits.append(float(snake.fruit))
                kills.append(float(snake.kills))
                self._update_grid(snake)
            dones.append(not snake.alive)

        # Generate fruit and add to the grid
        xs, ys = self._generate_fruits(fruit_taken)
        if xs is not None:
            self.grid[xs, ys] = Cell.FRUIT.value

        obs = self._get_obs()

        # for s_idx, rew in enumerate(rews):
        #     self.epi_scores[s_idx] += rew
        done_mask = 1. - np.asarray(dones)
        self.epi_scores = self.epi_scores + done_mask * np.asarray(rews)
        self.epi_steps = self.epi_steps + done_mask * np.ones(len(dones))
        self.epi_fruits = self.epi_fruits + done_mask * np.asarray(fruits)
        self.epi_kills = self.epi_kills + done_mask * np.asarray(kills)

        info = {}
        self.episode_length += 1
        if self.episode_length >= self.max_episode_steps:
            dones = [True] * self.num_snakes

        if self._done_fn(dones):
            sorted_scores = np.unique(np.sort(self.epi_scores)[::-1])
            ranks = np.array([0 for _ in range(self.num_snakes)])
            base_rank = 1
            for score in sorted_scores[::-1]:
                idx = np.where(np.array(self.epi_scores) == score)[0]
                ranks[idx] = base_rank
                base_rank += len(idx)
            info['rank'] = list(ranks)

            info.update(
                {'episode_scores': self.epi_scores,
                 'episode_steps': self.epi_steps,
                 'episode_fruits': self.epi_fruits,
                 'episode_kills': self.epi_kills})

            self._reset_epi_stats()

        return np.array(obs, dtype=np.uint8), rews, dones, info

    def _done_fn(self, dones):
        return all(dones)

    def save_gif(self, fp=None):
        if fp is None:
            curr_dir = os.getcwd()
            save_dir = os.path.join(curr_dir, 'tmp')
            now = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
            fp = os.path.join(save_dir, '{}.gif'.format(now))
            os.makedirs(save_dir, exist_ok=True)
        if not self.frame_buffer:
            warnings.warn(
                "You must call render('gif') first. No images to save."
            )
        else:
            print('Saving image to {}'.format(fp))
            self.frame_buffer[0].save(fp, save_all=True,
                                      append_images=self.frame_buffer[1:],
                                      format='GIF',
                                      loop=0)
        return fp

    def _reset_epi_stats(self):
        self.epi_scores = np.zeros(self.num_snakes)
        self.epi_steps = np.zeros(self.num_snakes)
        self.epi_fruits = np.zeros(self.num_snakes)
        self.epi_kills = np.zeros(self.num_snakes)

    def _init_obs(self):
        if self.image_obs:
            self.obs = deque(maxlen=self.frame_stack)
            for _ in range(self.frame_stack):
                self.obs.append(rgb_from_grid(self.grid, Cell, CellColors))
            obs = [np.concatenate(list(self.obs), axis=-1)
                   for _ in range(self.num_snakes)]
        else:
            self.obs = deque(maxlen=self.frame_stack)
            _obs = self._encode(self.grid, vision_range=self.vision_range)
            for _ in range(self.frame_stack):
                self.obs.append(_obs)
            obs = list(zip(*list(self.obs)))
            obs = [np.concatenate(o, axis=-1) for o in obs]

        return obs

    def _get_obs(self):
        if self.image_obs:
            self.obs.append(rgb_from_grid(self.grid, Cell, CellColors))
            obs = [np.concatenate(list(self.obs), axis=-1)
                   for _ in range(self.num_snakes)]
        else:
            _obs = self._encode(self.grid, vision_range=self.vision_range)
            self.obs.append(_obs)
            obs = list(zip(*list(self.obs)))
            obs = [np.concatenate(o, axis=-1) for o in obs]

        return obs

    def _encode(self, obs, vision_range=None):
        # Encode the observation. obs is self.grid
        # Returns the obs in HxWxC
        # May be overriden for customized observation
        env_objs = np.zeros([*obs.shape, 2], dtype=np.float32)
        snake_objs = np.zeros([*obs.shape,
                               3 * 2, self.num_snakes], dtype=np.float32)
        for r in range(obs.shape[0]):
            for c in range(obs.shape[1]):
                cell_value = obs[r, c]
                if cell_value in (Cell.WALL.value, Cell.FRUIT.value):
                    env_objs[r, c, cell_value - 1] = 1
                elif cell_value != Cell.EMPTY.value:
                    snake_id = cell_value // 10
                    obj_id = cell_value % 10
                    myself = np.zeros(self.num_snakes)
                    myself[snake_id] = 1
                    snake_objs[r, c, obj_id, :] = myself
                    snake_objs[r, c, obj_id - 3, :] = 1 - myself
        encoded_obs = [np.concatenate([env_objs, snake_objs[..., idx]],
                                      axis=-1)
                       for idx in range(self.num_snakes)]

        if vision_range:
            cropped_encoded_obs = []
            for full_obs in encoded_obs:
                head_pos = np.unravel_index(full_obs[:, :, 5].argmax(),
                                            full_obs[:, :, 5].shape)
                head_pos = np.array(head_pos)
                crop_range_min = np.maximum(head_pos - vision_range, 0)
                crop_range_max = np.minimum(head_pos + vision_range,
                                            np.array(obs.shape) - 1)
                cropped_obs = np.zeros((vision_range * 2 + 1,
                                        vision_range * 2 + 1,
                                        full_obs.shape[-1]))
                start = crop_range_min - head_pos + vision_range
                end = crop_range_max - head_pos + vision_range
                cropped_full_obs = full_obs[
                                   crop_range_min[0]:crop_range_max[0] + 1,
                                   crop_range_min[1]:crop_range_max[1] + 1, :]
                cropped_obs[start[0]:end[0] + 1,
                start[1]:end[1] + 1, :] = cropped_full_obs
                cropped_encoded_obs.append(cropped_obs)
            encoded_obs = cropped_encoded_obs

        return encoded_obs

    def _check_collision(self, next_head_coords):
        # Check for head, body, wall, fruit collision and assign new status
        dead_idxes = []
        fruit_idxes = []
        fruit_taken = 0
        for coord, idxes in next_head_coords.items():
            cell_value = self.grid[coord] % 10
            # Head collision or clear death
            if len(idxes) > 1 or cell_value in (Cell.WALL.value,
                                                Cell.BODY.value,
                                                Cell.HEAD.value):
                # if len(idxes) > 1:
                #     print("Death by collision", idxes)
                dead_idxes.extend(idxes)
                if cell_value == Cell.FRUIT.value:
                    fruit_taken += 1
                if cell_value in (Cell.BODY.value, Cell.HEAD.value):
                    self.snakes[self.grid[coord] // 10].kills += 1
            elif len(idxes) == 1 and cell_value == Cell.FRUIT.value:
                fruit_idxes.extend(idxes)
                fruit_taken += 1
        dead_idxes = list(set(dead_idxes))

        return dead_idxes, fruit_idxes, fruit_taken

    def _update_grid(self, snake):
        # Update the grid according to a snake's status
        if snake.alive:
            snake_id = 10 * snake.idx  # For distinguishing snakes in grid
            self.grid[snake.head_coord] = Cell.BODY.value + snake_id
            # could be current or prev if ate fruit
            prev_tail_coord = snake.move()
            if prev_tail_coord:
                # Didn't eat the fruit
                if self.grid[prev_tail_coord] == Cell.TAIL.value + snake_id:
                    self.grid[prev_tail_coord] = Cell.EMPTY.value
            self.grid[snake.head_coord] = Cell.HEAD.value + snake_id

            self.grid[snake.tail_coord] = Cell.TAIL.value + snake_id
        else:
            coord_list = snake.coords
            if self.grid[snake.coords[-1]] // 10 != snake.idx:
                coord_list = coord_list[:-1]
            if draw(self.grid, coord_list, Cell.EMPTY.value) is False:
                print('draw failed')
            snake.move()

    def _clear_overlap(self, list_of_coords):
        # Check for overlap in given snake coordinates
        flat_list = []
        for element in list_of_coords:
            flat_list.extend(element)
        unique_list = list(set(flat_list))
        return len(unique_list) == len(flat_list)

    def _generate_snakes(self):
        # Depth-first-search through the grid for possible snake positions
        candidates = dfs_sweep_empty(self.grid, self.snake_length)
        while True:
            # Randomly select init snake poses untill no overlap
            sample_idx = np.random.permutation(
                len(candidates)
            )[:self.num_snakes]
            samples = [candidates[si] for si in sample_idx]
            if self._clear_overlap(samples):
                break
        snakes = [Snake(idx, coords) for idx, coords in enumerate(samples)]

        return snakes

    def _generate_fruits(self, num_fruits=1):
        xs = ys = None
        if num_fruits:
            xs, ys = random_empty_coords(self.grid, num_coords=num_fruits)

        return xs, ys

    def _next_direction(self, direction, action):
        """
        0 == noop
        1 == left
        2 == right
        Change direction by +-90 degrees
        """
        angle = math.atan2(direction.value[1], direction.value[0])
        new_coord = (int(math.cos(angle + self.action_angle_dict[action])),
                     int(math.sin(angle + self.action_angle_dict[action])))
        return Direction(new_coord)

    def _next_direction_global(self, direction, action):
        """
        0 == noop
        1 == left
        2 == right
        3 == down
        4 == up
        Change direction to given global direction
        """
        new_direction = direction
        if direction.value[0] == 0:
            # snake is moving up or down
            if action == 3:
                new_direction = Direction.DOWN
            elif action == 4:
                new_direction = Direction.UP
        elif direction.value[1] == 0:
            # snake is moving lfet or right
            if action == 1:
                new_direction = Direction.LEFT
            elif action == 2:
                new_direction = Direction.RIGHT
        return new_direction